@@ -25,6 +25,7 @@
 
 #include "basic.h"
 #include "ptr.h"
+#include "optional.hpp"
 
 #include <iosfwd>
 #include <iterator>
@@ -74,9 +75,13 @@ class const_postorder_iterator;
 class symbol;
 struct symbolhasher;
 using symbolset = std::unordered_set<symbol,symbolhasher>;
+class wildcard;
+struct wildhasher;
+using wildset = std::unordered_set<wildcard,wildhasher>;
 using expairvec = std::vector<std::pair<ex,ex>>;
 using ocvector = std::vector<numeric>;
 using power_ocvector_map = std::map<ex, ocvector, GiNaC::ex_is_less>;
+using opt_ex = nonstd::optional<ex>;
 
 /** Lightweight wrapper for GiNaC's symbolic objects.  It holds a pointer to
  *  the other object in order to do garbage collection by the method of
@@ -150,11 +155,13 @@ public:
 
 	// operand access
 	size_t nops() const { return bp->nops(); }
+        size_t treesize() const;
 	size_t nsymbols() const;
         bool get_first_symbol(ex &x) const;
         symbolset symbols() const;
         symbolset free_symbols() const;
         std::unordered_set<unsigned> functions() const;
+        wildset wilds() const;
 	const ex op(size_t i) const { return bp->op(i); }
 	ex sorted_op(size_t i) const;
 	ex operator[](const ex & index) const { return (*bp)[index]; }
@@ -172,14 +179,18 @@ public:
 	ex imag_part() const { return bp->imag_part(); }
 
 	// pattern matching
-	bool has(const ex & pattern, unsigned options = 0) const { return bp->has(pattern, options); }
+	bool has(const ex & pattern, unsigned options = 0) const
+            { return bp->has(pattern, options); }
 	bool find(const ex & pattern, lst & found) const;
 	bool match(const ex & pattern) const;
-	bool match(const ex & pattern, lst & repl_lst) const { return bp->match(pattern, repl_lst); }
+	bool match(const ex & pattern, lst & repl_lst) const;
+	bool match(const ex & pattern, exmap& map) const
+            { return bp->match(pattern, map); }
+	bool match(const ex & pattern, exvector& vec) const;
+        std::vector<exmap> all_matches(const ex & pattern) const;
 
 	// substitutions
-	ex subs(const exmap & m, unsigned options = 0) const
-                { return bp->subs(m, options); }
+	ex subs(const exmap & m, unsigned options = 0) const;
 	ex subs(const lst & ls, const lst & lr, unsigned options = 0) const;
 	ex subs(const ex & e, unsigned options = 0) const;
 
@@ -222,7 +233,7 @@ public:
 #ifdef PYNAC_HAVE_LIBGIAC
         const giac::polynome to_polynome(ex_int_map& map, exvector& revmap) const;
 #endif
-        const CanonicalForm to_canonical(ex_int_map& map,
+        const CanonicalForm to_canonical(ex_int_umap& map,
                         power_ocvector_map& pomap, exvector& revmap) const;
         void collect_powers(power_ocvector_map& pomap) const;
 	ex numer() const;
@@ -252,6 +263,9 @@ public:
 	bool is_zero() const;
         bool is_one() const;
         bool is_minus_one() const;
+        bool is_negative_or_minus() const;
+        bool is_num_integer() const;
+        bool is_num_fraction() const;
 	
 	// noncommutativity
 	unsigned return_type() const { return bp->return_type(); }
@@ -259,7 +273,6 @@ public:
 
 	long gethash() const { return bp->gethash(); }
 
-private:
 	static ptr<basic> construct_from_basic(const basic & other);
 	static basic & construct_from_int(int i);
 	static basic & construct_from_pyobject(PyObject* o);
@@ -277,10 +290,13 @@ private:
 
 // member variables
 
-private:
 	mutable ptr<basic> bp;  ///< pointer to basic object managed by this
 };
 
+struct ex_hash {
+    long operator()(const ex& e) const { return e.gethash(); }
+};
+
 
 // performance-critical inlined method implementations
 
